#include <iostream>
#include <string>
#include <fstream>

// Erstes Tutorial fr OpenGL 3.3+
// Erstellt 2014 fr www.delphigl.com von hobbyfrickler und uebersetzt nach C++ von glAwesome.
// Die SDL2-Header sind auf www.libsdl.org erhltlich: https://www.libsdl.org/download-2.0.php
// Der beigelegte gl3w-Header stammt von hier (und wurde minimal modifiziert): https://github.com/shakesoda/gl3w

// Es wird ein zweidimensionales Dreieck mit Farbverlauf gezeichnet.
// Benutzte Techniken: VBO, VAO, simpler Vertex- & Fragmentshader,
// Fensterverwaltung und Ereignisbehandlung mittels SDL2.

// Das Verzeichnis, in dem sich der Ordner SDL2 mit den Headerdateien befindet, muss dem Compiler
// bekannt gegeben werden. Siehe auch "Tutorial SDL2 Installation":
// http://wiki.delphigl.com/index.php/Tutorial_SDL2_Installation
#include <SDL2/SDL.h>

// Als OpenGL-Loader setzen wir gl3w ein. Wer sich die gl3w-Dateien selbst bauen will, kann sich
// das dafuer noetige Python-Skript hier herunterladen: https://github.com/shakesoda/gl3w
#include "gl3w/gl3w.h"

// Globale Variablen sind _kein_ schner Stil, den man sich angewoehnen sollte.
// Hier wird davon Gebrauch gemacht, um den Code moeglichst simpel zu halten.
SDL_Window* g_Window;
SDL_GLContext g_Context;
bool g_Running = true;
GLuint g_VAO = 0;
GLuint g_VBO[2] = {0,0};

bool Setup_SDL2()
{
  // Initialisieren des SDL-Videosubsystems
  if(SDL_Init(SDL_INIT_VIDEO) != 0)
  {
    std::cout << "Initializieren des SDL-Videosubsystems fehlgeschlagen: " << SDL_GetError() << std::endl;
    return false;
  }
  // Ein Fenster in der Mitte des Desktops erstellen, Fenstergre ist
  // 640x480 und vernderbar, OpenGL-Untersttzung ist aktiviert
  g_Window = SDL_CreateWindow("OpenGL 3.3 Tutorial 1", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, 640, 480, SDL_WINDOW_OPENGL|SDL_WINDOW_RESIZABLE);
  if(g_Window == 0)  // Fenster korrekt erstellt?
  {
    std::cout << "SDL-Fenster konnte nicht erstellt werden: " << SDL_GetError() << std::endl;
    return false;
  }
  // Die gewnschte Minimum OpenGL-Version festlegen - wir benutzen OpenGL 3.3
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3);
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 3);
  g_Context = SDL_GL_CreateContext(g_Window);  // OpenGL-Kontext erstellen
  if(g_Context == 0)  // Kontext korrekt erstellt?
  {
    std::cout << "OpenGL-Kontext konnte nicht erstellt werden: " << SDL_GetError() << std::endl;
    return false;
  }
  return true;
}

void SetViewport_OpenGL(int w, int h)
{
  // Wird beim ndern der Fenstergre aufgerufen, passt den OpenGL-Viewport
  // an die neue Fenstergre an.
  glViewport(0, 0, w, h);
  // Anmerkung: Nach nderung der Fenstergre sehen die Objekte evtl. perspektivisch
  // verzerrt aus. Das liegt daran, dass wir noch keine Projektionsmatrix benutzen.
}

void HandleEvents_SDL()
{
  SDL_Event event;
  while(SDL_PollEvent(&event) > 0)
  {
    switch(event.type)
    {
      case SDL_WINDOWEVENT:
      { // Ereignis betrifft SDL-Fenster
        switch(event.window.event)
        {
          case SDL_WINDOWEVENT_SIZE_CHANGED:
          { // Fenstergre wurde verndert
            // In event.window.data1 & data2 stehen verschiedene Daten, abhngig
            // von der Art des Ereignisses, hier:
            // event.window.data1 = neue Fensterbreite, data2 = neue Fensterhhe
            SetViewport_OpenGL(event.window.data1, event.window.data2);
            break;
          }
          case SDL_WINDOWEVENT_CLOSE:
          { // Fenster soll geschlossen werden, z.B. Klick auf Fensterschliesymbol
            g_Running = false; // Hauptschleife soll beendet werden
            break;
          }
        }
      }
    }
  }
}

// Funktion um Fehler von OpenGL-Objekten auszulesen, momentan knnen Kompilier-
// und Linkerfehler von Shadern ausgelesen werden.
std::string GetInfoLogMessage(GLuint glObject)
{
  std::string str;
  int length = 0;
  if(glIsProgram(glObject))
  {
    // Lnge der Meldung im InfoLog auslesen
    glGetProgramiv(glObject, GL_INFO_LOG_LENGTH, &length);
    // Gre des Ergebnisstrings anpassen
    str.resize(length);
    // Meldung in Ergebnisstring kopieren
    glGetProgramInfoLog(glObject, length, 0, &str[0]);
  }
  else if(glIsShader(glObject))
  {
    glGetShaderiv(glObject, GL_INFO_LOG_LENGTH, &length);
    str.resize(length);
    glGetShaderInfoLog(glObject, length, 0, &str[0]);
  }
  return str;
}

// Gibt fuer jeden Shader und Program das entsprechende Log aus
std::string GetInfoLog(GLuint hVertShader, GLuint hFragShader, GLuint hProgram)
{
  std::string logStr;
  logStr += "Vertex Shader\n==============\n";
  logStr += GetInfoLogMessage(hVertShader);

  logStr += "\n\nFragment Shader\n================\n";
  logStr += GetInfoLogMessage(hFragShader);

  logStr += "\n\nProgram\n========\n";
  logStr += GetInfoLogMessage(hProgram);
  return logStr;
}

// Liest eine Textdatei und gibt den Inhalt als string zurueck
std::string LoadShaderFromFile(const std::string& filename)
{
  std::string str;
  std::ifstream file(filename.c_str());
  if(!file.is_open())
  {
    std::cout << "Shaderdatei nicht gefunden: " << filename << std::endl;
    return str;
  }
  // Dateigroesse herausfinden
  file.seekg(0, std::ios_base::end);
  int fileSize = file.tellg();
  file.seekg(0, std::ios_base::beg);
  // Den noetigen Speicherplatz reservieren und mit dem Terminierungszeichen \0 fuellen
  str.resize(fileSize+1, '\0');
  file.read(&str[0], fileSize);
  return str;
}

// Laedt unsere beiden Shader, kompiliert und linkt sie
GLuint LoadShaders(const std::string& vsFilename, const std::string& fsFilename)
{
  // Dateien laden
  std::string strVertShader = LoadShaderFromFile(vsFilename);
  std::string strFragShader = LoadShaderFromFile(fsFilename);
  if(strVertShader == "" || strFragShader == "") // Konnte eine der Dateien nicht geladen werden?
    return 0;  // 0 ist kein gueltiges Prgram-Objekt

  // Objekte erzeugen
  GLuint hProgram = glCreateProgram();
  GLuint hVertShader = glCreateShader(GL_VERTEX_SHADER);
  GLuint hFragShader = glCreateShader(GL_FRAGMENT_SHADER);

  // OpenGL den Quelltext uebergeben
  const int vsLen = strVertShader.length();
  const int fsLen = strFragShader.length();
  const char* cstrVertShader = strVertShader.c_str();
  const char* cstrFragShader = strFragShader.c_str();
  glShaderSource(hVertShader, 1, &cstrVertShader, &vsLen);
  glShaderSource(hFragShader, 1, &cstrFragShader, &fsLen);

  // Shader kompilieren
  glCompileShader(hVertShader);
  glCompileShader(hFragShader);

  // Gab es Fehler beim kompilieren?
  int vsStatus, fsStatus, progStatus;
  glGetShaderiv(hVertShader, GL_COMPILE_STATUS, &vsStatus);
  glGetShaderiv(hFragShader, GL_COMPILE_STATUS, &fsStatus);

  // linken
  glAttachShader(hProgram, hVertShader);
  glAttachShader(hProgram, hFragShader);
  glLinkProgram(hProgram);

  // InfoLog ausgeben
  std::cout << GetInfoLog(hVertShader, hFragShader, hProgram) << std::endl;

  // Shader zum loeschen markieren (sie werden dann geloscht, sobald hProgram freigegeben wird)
  glDeleteShader(hVertShader);
  glDeleteShader(hFragShader);

  // Linkerfehler?
  glGetProgramiv(hProgram, GL_LINK_STATUS, &progStatus);
  if(vsStatus == GL_FALSE || fsStatus == GL_FALSE || progStatus == GL_FALSE)
  {
    std::cout << "Fehler beim Kompilieren oder Linken!\n";
    glDeleteProgram(hProgram);
    return 0;
  }

  glUseProgram(hProgram);
  return hProgram;
}

// Erzeugt die Vertexdaten und ldt sie in den Grafikspeicher
void CreateScene()
{
  const float positionData[9] = {-0.7,-0.5,-1.0,  0.7,-0.5,-1.0,  0.0, 0.7,-1.0};
  const float colorData[9]    = { 1.0, 0.0, 0.0,  0.0, 0.0, 1.0,  0.0, 1.0, 0.0};

  glGenVertexArrays(1, &g_VAO);  // Das VAO erstellen
  glGenBuffers(2, &g_VBO[0]);  // Zwei VBO's erstellen

  glBindVertexArray(g_VAO);  // Aktivieren des VAO
  // Aktivieren des ersten VBO im VAO = Positionsdaten des Dreiecks
  glBindBuffer(GL_ARRAY_BUFFER, g_VBO[0]);
  // VBO mit Daten aus Array befllen und an GPU schicken
  glBufferData(GL_ARRAY_BUFFER, sizeof(positionData), positionData, GL_STATIC_DRAW);
  // Wir wollen das erste Vertexattribut (inPosition) "erklren"
  glEnableVertexAttribArray(10);
  // Hier teilt man OpenGL mit wie die Daten im VBO zu interpretieren sind
  // 0 = Index des Attributs (im Vertexshader mittels "location" festgelegt),
  // 3 = Anzahl der Elemente pro Vertex (x,y,z),
  // GL_FLOAT = Datentyp der Daten,
  // FALSE = Kein normalisieren ausfhren (Daten werden nicht verndert),
  // 12 = Byte-Offset zwischen 2 Attributen (3 * sizeof(GLfloat)), in diesem Beispiel knnte man auch 0 angeben (weil alle Elemente direkt aufeinanderfolgen)
  // Pointer(0) = Zeiger auf erstes zu benutzendes Element im Buffer
  glVertexAttribPointer(10, 3, GL_FLOAT, false, 3*sizeof(positionData[0]), 0);

  // Aktivieren des zweiten VBO im VAO = Farbdaten Dreieck
  glBindBuffer(GL_ARRAY_BUFFER, g_VBO[1]);
  glBufferData(GL_ARRAY_BUFFER, sizeof(colorData), colorData, GL_STATIC_DRAW);
  glEnableVertexAttribArray(11);  // Wir wollen das zweite Vertexattribut (inColor) "erklren"
  glVertexAttribPointer(11, 3, GL_FLOAT, false, 3*sizeof(colorData[0]), 0);
}

void RenderScene()
{
  // Letztes Bild und Tiefenpuffer lschen
  glClear(GL_COLOR_BUFFER_BIT);
  glBindVertexArray(g_VAO);  // Aktivieren des VAO
  // Zeichnen des Dreiecks mittels VBOs (im VAO), GL_TRIANGLES bewirkt das Dreiecke erstellt
  // werden, 0 = Startindex im VBO, 3 = Anzahl der zu rendernden Vertices
  glDrawArrays(GL_TRIANGLES, 0, 3);
  // Front- und Backbuffer tauschen
  SDL_GL_SwapWindow(g_Window);
}

int main(int argc, char* args[])
{
  if(!Setup_SDL2())  // Fenster und Kontext erstellen
    return -1;
  if(gl3wInit() != 0)  // OpenGL initialisieren
  {
    std::cout << "gl3w konnte nicht initialisiert werden!\n";
    return -1;
  }

  SetViewport_OpenGL(640, 480);  // Viewport setzen
  GLuint hProgram = LoadShaders("shader/VertexShader.txt",    // Shader laden
                                "shader/FragmentShader.txt");
  CreateScene();  // Szene erstellen

  // Hauptschleife des Programms
  while(g_Running && g_Window != 0)
  {
    HandleEvents_SDL();
    RenderScene();
  }

  // OpenGL Ressourcen freigeben
  glDeleteBuffers(2, &g_VBO[0]);
  glDeleteVertexArrays(1, &g_VAO);
  if(hProgram != 0)
    glDeleteProgram(hProgram);

  // SDL Ressourcen freigeben
  SDL_GL_DeleteContext(g_Context);
  SDL_DestroyWindow(g_Window);
  SDL_Quit();
  return 0;
}
